/*

  Copyright 2010, 2011 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.app.stage.example; 

import cf.api.ApplicationKnowledge;
import cf.api.CyberApp;
import cf.api.CyberApplication;
import cf.api.CyberNode;
import cf.api.devices.Area;
import cf.api.devices.Camera;
import cf.api.devices.CameraApplication;
import cf.api.devices.Image;
import cf.api.devices.Laser;
import cf.api.devices.LaserApplication;
import cf.api.devices.LaserSample;
import cf.api.devices.MotionSensor;
import cf.api.devices.MotionSensorApplication;
import cf.api.devices.NoiseSensor;
import cf.api.devices.NoiseSensorApplication;
import cf.api.devices.Position;
import cf.api.devices.PositionApplication;
import cf.api.devices.Powerpack;
import cf.api.devices.PowerpackApplication;
import cf.sim.stage.devices.SimPositionStage;

class PositionKnowledge extends ApplicationKnowledge
{
  String robot;
  Area area;
  PositionKnowledge(CyberApplication creator)
  {
    super(creator);
  }
  public String logRepLong()
  {
    return super.logRepLong() + " Robot: " + robot + ", "+ area.logRepLong();
  }
}

class LaserKnowledge extends ApplicationKnowledge
{
  String robot;
  boolean stop, avoid;
  boolean left, right;
  LaserKnowledge(CyberApplication creator)
  {
    super(creator);
  }
  public String logRepLong()
  {
    return super.logRepLong() + " Robot: " + robot + " stop[" + stop + "] , avoid[" + avoid +"] ";
  }
  public void setLeft()
  {
    left = true;
    right = false;
  }
  public void setRight()
  {
    left = false;
    right = true;
  }
}

class PowerKnowledge extends ApplicationKnowledge
{
  String robot;
  double stored, dissipated;
  PowerKnowledge(CyberApplication creator)
  {
    super(creator);
  }
  public String logRepLong()
  {
    return super.logRepLong() + " Robot: " + robot + " " + "Stored: "+ stored + ", Dissipated: " + dissipated;
  }
}

class MotionKnowledge extends ApplicationKnowledge
{
  String robot;
  Area area;

  MotionKnowledge(CyberApplication creator)
  {
    super(creator);
  }

  public String logRepLong()
  {
    return super.logRepLong() + " Robot: " + robot;
  }
}

class NoiseKnowledge extends ApplicationKnowledge
{
  String robot;
  Area area;

  NoiseKnowledge(CyberApplication creator)
  {
    super(creator);
  }

  public String logRepLong()
  {
    return super.logRepLong() + " Robot: " + robot;
  }
}

class ImageKnowledge extends ApplicationKnowledge
{
  String robot;
  Area area;
  Image image;

  ImageKnowledge(CyberApplication creator)
  {
    super(creator);
  }

  public boolean subsumed(ApplicationKnowledge k)
  {
    return k.creationTime() == creationTime(); // no subsumption
  }

  public String logRepLong()
  {
    return super.logRepLong() + " Robot: " + robot + " Image: " + image.logRepLong();
  }
}

class RootApp extends CyberApp
{
  public RootApp(CyberNode node)
  {
    super(node);
  }

  public void handleKnowledge(ApplicationKnowledge k)
  {
    if(k instanceof ImageKnowledge) handleImage((ImageKnowledge)k);
  }

  void handleImage(ImageKnowledge kimage)
  {
    log("handling image");
  }
}

class LaserApp extends CyberApp implements LaserApplication
{
  Laser laser;
  double minfrontdistance = 1.0;
  double stopdist = 0.5;
  boolean obstruction = false;
  boolean stop = false;
  double minleft = Double.POSITIVE_INFINITY;
  double minright = Double.POSITIVE_INFINITY;

  public LaserApp(CyberNode node, Laser laser)
  {
    super(node);
    this.laser = laser;
    laser.addApplication(this);
  }

  public void handleLaserUpdate()
  {
    LaserSample [] ls = laser.getSamples();
    if(ls !=null) 
    {
      int sample_count = ls.length;
      //System.out.println("Sample Count : " + sample_count);
      obstruction = false; stop = false;
      for (int i = 0; i < sample_count; i++)
      {
        if( (i > (sample_count/5)*2)
            && (i < (sample_count - (sample_count/5)*2))
            && ls[i].getRange() < minfrontdistance)
        {
          //log( "  obstruction!" );
          //System.out.println( "  obstruction!" + ls[i].getRange());
          obstruction = true;
        }

        if( ls[i].getRange() < stopdist )
        {
          //log( "  stopping!" );
          //System.out.println( "  stopping!"+ ls[i].getRange() );
          stop = true;
        }

        if( i > sample_count/2 )
          minleft = Math.min( minleft, ls[i].getRange());
        else
          minright = Math.min( minright, ls[i].getRange() );
      }
      if(obstruction || stop) 
      {
        LaserKnowledge lk = new LaserKnowledge(this);
        lk.stop = stop;
        lk.avoid = obstruction;
        lk.robot = laser.parentname();
        if(minright < minleft)
          lk.setLeft();
        else
          lk.setRight();
        postKnowledge(lk);
        System.out.println("Post Laser Knowledge");
      }
    }
  }
}

class MotionSensorApp extends CyberApp implements MotionSensorApplication
{    
  MotionSensor sensor;

  public MotionSensorApp(CyberNode node, MotionSensor sensor)
  {
    super(node);
    this.sensor = sensor;
    sensor.addApplication(this);
  }

  public void handleMotion()
  {
    log("handling motion");
    MotionKnowledge kmotion = new MotionKnowledge(this);
    kmotion.robot = sensor.parentname();
    postKnowledge(kmotion);
        System.out.println("Post motion Knowledge");
  }
}

class NoiseSensorApp extends CyberApp implements NoiseSensorApplication
{
  NoiseSensor sensor;

  public NoiseSensorApp(CyberNode node, NoiseSensor sensor)
  {
    super(node);
    this.sensor = sensor;
    sensor.addApplication(this);
  }

  public void handleNoise()
  {
    log("handling noise");
    NoiseKnowledge knoise = new NoiseKnowledge(this);
    knoise.robot = sensor.parentname();
    postKnowledge(knoise);
        System.out.println("Post noise Knowledge");
  }
}

class CameraApp extends CyberApp implements CameraApplication
{
  Camera camera;

  Area area;

  public CameraApp(CyberNode node, Camera camera)
  {
    super(node);
    this.camera = camera;
    camera.addApplication(this);
  }

  public void handleKnowledge(ApplicationKnowledge knew)
  {
    //System.out.println(name()+  " : " + knew.logRepShort());
    if(knew instanceof MotionKnowledge) handleMotionKnowledge((MotionKnowledge)knew);
    else if(knew instanceof NoiseKnowledge) handleNoiseKnowledge((NoiseKnowledge)knew);
    else if(knew instanceof PositionKnowledge) handlePositionKnowledge((PositionKnowledge)knew);
  }

  void handlePositionKnowledge(PositionKnowledge pk) 
  {
    log("handling position knowledge");
    if(pk.robot.equals(camera.parentname()))
    {
      Image image = camera.takeImage( pk.area );
      ImageKnowledge kimage = new ImageKnowledge(this);
      kimage.robot = camera.parentname();
      kimage.area = pk.area;
      kimage.image = image;
      postKnowledge(kimage);
	  System.out.println("Post image k");
    }
  }

  void handleMotionKnowledge(MotionKnowledge kmotion)
  {	
    log("handling motion knowledge");

    if(kmotion.robot.equals(camera.parentname()) && kmotion.area == area)
    {
      Image image = camera.takeImage(area);
      ImageKnowledge kimage = new ImageKnowledge(this);
      kimage.robot = camera.parentname();
      kimage.area = area;
      kimage.image = image;
      postKnowledge(kimage);
	  System.out.println("Post image k");
    }
  }

  void handleNoiseKnowledge(NoiseKnowledge knoise)
  {	
    log("handling noise knowledge");

    if(knoise.robot.equals(camera.parentname()) && knoise.area == area)
    {
      Image image = camera.takeImage(area);
      ImageKnowledge kimage = new ImageKnowledge(this);
      kimage.robot = camera.parentname();
      kimage.area = area;
      kimage.image = image;
      postKnowledge(kimage);
	  System.out.println("Post image k");
    }
  }

}

class PowerpackApp extends CyberApp implements PowerpackApplication
{
  Powerpack powerpack;

  public PowerpackApp(CyberNode node, Powerpack powerpack)
  {
    super(node);
    this.powerpack = powerpack;
    powerpack.addApplication(this);
  }
  
  public void handlePowerpackUpdate()
  {
  }

  public void handleKnowledge(ApplicationKnowledge knew)
  {
    //System.out.println(name()+  " : " + knew.logRepShort());
    if(knew instanceof ImageKnowledge) handleImage((ImageKnowledge)knew);
  }

  public void handleImage(ImageKnowledge ik)
  {

    log("handling image knowledge");
    if(ik.robot.equals(powerpack.parentname()))
    {
    PowerKnowledge pk = new PowerKnowledge(this);
    pk.stored =  powerpack.GetStored();
    pk.dissipated = powerpack.GetDissipated();
    pk.robot = powerpack.parentname();
    postKnowledge(pk);
	  System.out.println("Post pp k");
    }
  }

}

class PositionApp extends CyberApp implements PositionApplication
{
  Position position;
  //private JavaStageCtrl ctrl;
  private boolean is_sent=false;
  private double avoidspeed = 0.05;
  private int avoidcount = 0;
  private int avoidduration = 10;
  private double avoidturn = 0.5;
  private double cruisespeed = 0.4;

  public PositionApp(CyberNode node, int node_idx, Position position)
  {
    super(node);
    this.position = position;
    position.addApplication(this);
    //this.ctrl = ((SimPositionStage)position).ctrl;

    // set initial robot position based on node index
    //double x,y;
    //double width, height;
    //int node_idx = ctrl.GetNodeIndex();
    //width = SimulationInterface._GetFloorplanSize(ctrl.GetRobotP()).x;
    //height = SimulationInterface._GetFloorplanSize(ctrl.GetRobotP()).y;
    //x = (node_idx*4) % (int)(width-2) - ((int)(width)/2-1) ;
    //y = - ( 2*(node_idx*4 / (int)(height-2)) - ((int)(height)/2-1) );

    //((SimPositionStage)position).setPos( x, y);

    //// move robot 0 to 0,0
    if(node_idx == 0) 
    {
      postEvent(1.0);
    }
  }

  public void handleEvent()
  {
    if(avoidcount>0) 
    {
      if(avoidcount == 10)
        ((SimPositionStage)position).SetSpeed(avoidspeed, 0, 0);
      if(avoidcount  == 9)
        ((SimPositionStage)position).SetTurnSpeed(avoidturn);
      if(avoidcount == 6)
        ((SimPositionStage)position).SetSpeed(0.25, 0, 0);
      avoidcount --;
    }
    else 
    {
      position.move(0,0);
      //System.out.println("LETs GO!");
    }
    postEvent(0.5);
  }

  public void handlePositionUpdate()
  {
    if(position.posx() < 1 && position.posx() > -1 &&
        position.posy() < 1 && position.posy() > -1 &&
        !is_sent) 
    {
      System.out.println("Arrive");
      PositionKnowledge pk = new PositionKnowledge(this);
      pk.robot = position.parentname();
      pk.area = new Area("Area Center" , position.posx(), position.posy() );
	  postKnowledge(pk);
      System.out.println("post position k");
      is_sent = true;
    }
  }

  public void handleKnowledge(ApplicationKnowledge k )
  {
    //System.out.println(name()+  " : " + k.logRepShort());
    if( k instanceof LaserKnowledge) handleLaser( (LaserKnowledge) k );
  }

  public void handleLaser(LaserKnowledge k)
  {
    if( k.robot.equals(position.parentname()))
    {
      if((k.avoid || k.stop) && avoidcount < 1)
      {
        //System.out.println("LASER KNOWLEDGE RECEIVED");
        if(k.stop) avoidspeed = 0.0;
        else avoidspeed = 0.05;
        
        if(k.left) avoidturn = 0.7;
        else avoidturn = -0.7;

        avoidcount = 10;
      }
    }
  }
}

class RobotApp extends CyberApp
{
  public RobotApp(CyberNode node)
  {
    super(node);
    //System.out.println("App: " + name()+" , time: " + time());
  }
  public void handleImage(ImageKnowledge ik)
  {
    //System.out.println(ik.logRepLong() + " "+ ik.area.logRepLong());
  }
  public void handleKnowledge(ApplicationKnowledge k)
  {
    //System.out.println(name()+  " : " + k.logRepShort());
    if(k instanceof ImageKnowledge) handleImage((ImageKnowledge)k);
  }
}
