/*

  Copyright 2010, 2011 SRI International and Seoul National University

  Licensed under the Apache License, Version 2.0 (the "License"); you may
  not use this file except in compliance with the License. You may obtain
  a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

  Unless required by applicable law or agreed to in writing, software
  distributed under the License is distributed on an "AS IS" BASIS,
  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
  See the License for the specific language governing permissions and
  limitations under the License.

  The development was supported by National Science Foundation Grant 0932397:
  A Logical Framework for Self-Optimizing Networked Cyber-Physical Systems

*/

package cf.sim.stage;

import java.util.LinkedList;

import cf.common.KnowledgeManager;
import cf.common.Model;
import cf.common.Param;
import cf.sim.SimEngine;
import cf.sim.SimNode;
import cf.sim.SimWorld;
import cf.sim.dtn.LinkType;
import cf.sim.dtn.SimLink;
import cf.sim.dtn.Transmission;
import cf.util.Dist;
import cf.util.Log;

public class JavaStageCtrlCyberFramework extends JavaStageCtrl
{

  public JavaStageCtrlCyberFramework(int robot_p, int node_index)
  {
    super(robot_p, node_index);
    //  WifiInterface._add_rcv_msg_fn(new WifiRcv(this), GetRobotP());
  }

  public JavaStageCtrlCyberFramework()
  {
    if(!Param.EnableStage)
    {
      System.err.println("Param.EnableStage should be on when using Stage");
      System.exit(1);
    }
  }

  private SimNode wifi_node;
  private static LinkedList<SimNode> allWifiNodes= new LinkedList<SimNode>();

  private static SimWorld world;

  public SimWorld world() 
  {
    return world;
  }

  /* called once by Stage world.cc before all robot's init() is called */
  public static void world_init_pre()
  {
    try{ SimEngine.beginRound(); } catch(Exception e) 
    {
      e.printStackTrace();
    }
    long seed = System.currentTimeMillis();
    SimEngine.dist = new Dist(seed);
    KnowledgeManager.dist = new Dist(seed);
    if(!Param.EnableStage) {
      System.err.println("EnableStage Should Be ON");
      System.exit(1);
    }
    if(!Param.UseSimNetwork) 
    {
      Param.UseStageLinks = false;
      Param.UseStageTransfer = false;
    }
    Param.UseRealNetwork = false;
    Param.RealTime = false;
    Param.SharedKnowledgeBase = false;
    Param.CyberThread = false;

    world = new SimWorld();
    SimEngine.log.println(1, "CyberFramework World created, World SIze: " + world.totModels());
  }

  /* called once per each robot in stage before running event Q */
  public void init()
  {
  }

  /* called once by Stage world.cc after all robot's init() */
  public static void world_init_after()
  {
    world.initialize();
    SimEngine.log.println(1, "World Size: " + world.totModels());
    initLinks();
    addRcvMsgFn();
  }

  private static void addRcvMsgFn()
  {
    int num_robot_p = SimulationInterface.GetNumRobotP();
    int [] robot_p_v = SimulationInterface.GetRobotPVector();
    for(int i=0;i<num_robot_p;i++) 
    {
      WifiInterface._add_rcv_msg_fn(new WifiRcv(), robot_p_v[i]);
    }
  }

  private static void initLinks()
  {
    /* add persistent linke btw parent & child */
    for(Model model : world.allModels() )
    {
      if(model instanceof SimNode)
      {
        Model p = model.parent();
        if(p != null) 
        {
          SimLink l = 
            world.addSymmetricLink((SimNode)(p), (SimNode)model, LinkType.ParentChild, Param.LinkCost, Param.LinkLatency, Param.LinkBandwidth, Param.LinkErrorRate);
          world.handleLinkChange(true, l);
        }
      }
    }

    /* add wifi links among robot nodes */

    for(SimNode me : allWifiNodes)
    {
      for(SimNode you : allWifiNodes)
      {
        if(me.index() != you.index())
        {
          if(!Param.UseStageLinks && Param.EnableStage && Param.UseSimNetwork)
          {
            SimLink l =
              world.addLink(me, you, LinkType.Wireless, Param.LinkCost, Param.LinkLatency, Param.LinkBandwidth, Param.LinkErrorRate);
            world.handleLinkChange(true, l);
          }
          else
          {
            world.addLink(me, you, LinkType.Wireless, Param.LinkCost, Param.LinkLatency, Param.LinkBandwidth, Param.LinkErrorRate);
          }
        }
      }
    }
  }

  /* called after stage simulation end */
  public static void world_wrapup()
  {
    try { SimEngine.endRound(); } catch (Exception e) 
    {
      e.printStackTrace();
    }
    SimEngine.rep.println("Final World State");
    SimNode.printAll(world);
  }

  /* fix the network node */
  public void setWifiNode(SimNode n)
  {
    wifi_node = n;
    allWifiNodes.add(n);
  }

  /* is called every link update time in Stage( model_wifi.cc ) */
  public void update_link()
  {
    if(Param.UseStageLinks && Param.EnableStage && Param.UseSimNetwork) 
    {
      if( wifi_node != null) 
      {
        SimNode me = wifi_node;
        // Get My Neighbors
        WifiNeighbor [] wn = WifiInterface._GetLinkInfo(GetRobotP()); 
        if( wn == null)
        {
          for(SimNode you : allWifiNodes)
          {
            for(SimLink l : me.outgoingLinks())
            {
              if(l.receiver().index() == you.index())
                world.handleLinkChange(false, l); // no neighbor , all link related wifi node down
            }
          }
          return;
        }

        // check neighbor relationship btw wifi nodes
        // if neighbor then link up
        for(SimNode you : allWifiNodes)
        {
          if(me.index() != you.index())
          {
            if(is_NodeInNeighbor(you, wn))
            {
              for(SimLink l : me.outgoingLinks())
              {
                if(l.receiver().index() == you.index())
                {
                  world.handleLinkChange(true, l);
                }
              }
            }
            else
            {
              for(SimLink l : me.outgoingLinks())
              {
                if(l.receiver().index() == you.index())
                {
                  world.handleLinkChange(false, l);
                }
              }
            }
          }
        }
      }
    }
  }

  private boolean is_NodeInNeighbor(SimNode node, WifiNeighbor [] wn)
  {
    if( wn == null) return false;
    for(WifiNeighbor n : wn)
    {
      if(n.node_index == node.ctrl().GetNodeIndex())
        return true;
    }
    return false;
  }

  // leave time log at log file
  public static  void log( )
  {
    SimEngine.log.println(1, "--- SimEngine Time " + Log.tf.format(SimEngine.time()) + " ---");
  }

}
